from .base import BaseDetector
from .hr_patchcore import HRPatchCore
from .hr_fastflow import HRFastFlow
from .hr_vitad import HRVitAD
from .hr_realnet import HRRealNet
from .hr_destseg import HRDesTSeg
from .hr_rdpp import HRRDPP
from .hr_padim import HRPaDiM
